#!/usr/bin/env python3

"""
soc_helper
Copyright (C) 2023-2024  M. Williges (spam at zut punkt de)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# System-Importe
import logging
import paho.mqtt.client as mqtt
import paho.mqtt.publish as publish
import json
import sys
import time
import importlib
from os.path import exists
import os
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

# Importe von lokalen Dateien
import configuration	#Konfiguration des soc_helper.py
import ladevorgangclass #Daten und Methoden des aktuellen Ladevorgangs

# rudimentäre Prüfung der Konfiguration
def checkConfig():
    # Prüfen, ob alle nötigen Parameter definiert sind
    if not hasattr(configuration, 'USE_SPRITMONITOR'):
        logging.error('USE_SPRITMONITOR ist undefiniert')
        sys.exit()
    elif not isinstance(configuration.USE_SPRITMONITOR, bool):
        logging.error('USE_SPRITMONITOR ist nicht "True" oder "False"')
        sys.exit()
    elif configuration.USE_SPRITMONITOR:
        # Prüfen, ob bei Spritmonitor-Nutzung ein BEARER-Token gesetzt ist
        bearer_token = os.environ.get("SPRITMONITOR_BEARER_TOKEN")
        if bearer_token == None:
            logging.error('Spritmonitor aktiv, aber Umgebungsvariable SPRITMONITOR_BEARER_TOKEN nicht gesetzt, siehe README.txt.')
            sys.exit()
        if not ( hasattr(configuration, 'SPRITMONITOR_VEHICLE_ID') and
                 hasattr(configuration, 'SPRITMONITOR_FUELSORT') and
                 hasattr(configuration, 'SPRITMONITOR_FUELPRICE') and
                 hasattr(configuration, 'SPRITMONITOR_ATTRIBUTES') ):
            logging.error('Spritmonitor aktiv, aber mindestens einer der folgenden Parameter ist undefiniert:')
            logging.error('SPRITMONITOR_VEHICLE_ID, SPRITMONITOR_FUELSORT,SPRITMONITOR_FUELPRICE, SPRITMONITOR_ATTRIBUTES')
            sys.exit()
    if not hasattr(configuration, 'SOC_SENDER_ID'):
        logging.error('SOC_SENDER_ID ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'SOC_RECV_ID'):
        logging.error('SOC_RECV_ID ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'DST_SENDER_ID'):
        logging.error('DST_SENDER_ID ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'DST_RECV_ID'):
        logging.error('DST_RECV_ID ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'SOC_REQUEST'):
        logging.error('SOC_REQUEST ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'DST_REQUEST'):
        logging.error('DST_REQUEST ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'OPENWB_IP'):
        logging.error('OPENWB_IP ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'OPENWB_SETSOC_TOPIC'):
        logging.error('OPENWB_SETSOC_TOPIC ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'OPENWB_GETSOC_TOPIC'):
        logging.error('OPENWB_GETSOC_TOPIC ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'OPENWB_CHARGECOUNTER_TOPIC'):
        logging.error('OPENWB_CHARGECOUNTER_TOPIC ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'OPENWB_GETPLUGSTATE_TOPIC'):
        logging.error('OPENWB_GETPLUGSTATE_TOPIC ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'OPENWB_FACTORCOUNTER'):
        logging.error('OPENWB_FACTORCOUNTER ist undefiniert')
        sys.exit()
    if not hasattr(configuration, 'WICAN_SER'):
        logging.error('WiCAN-Seriennummer WICAN_SER ist undefiniert')
        sys.exit()
    elif len(configuration.WICAN_SER)!=12:
        logging.error('WICAN_SER hat die falsche Länge (Muss 12 Hex-Ziffern sein)')
        sys.exit()
    
    # Prüfen der CAN-Konfiguration
    v = configuration.SOC_SENDER_ID
    if not isinstance(v, int):
        logging.error(f'SOC_SENDER_ID ist nicht vom Typ int.')
        sys.exit()
    if (v<0 or v>4095):
        logging.error(f'SOC_SENDER_ID hat den Wert {v} und liegt nicht zwischen 0 und 4095.')
        sys.exit()
    v = configuration.SOC_RECV_ID
    if not isinstance(v, int):
        logging.error(f'SOC_RECV_ID ist nicht vom Typ int.')
        sys.exit()
    if ( v < 0 or v > (2**30-1) ):
        logging.error(f'SOC_RECV_ID hat den Wert {v} und liegt nicht zwischen 0 und 4095.')
        sys.exit()
    if not isinstance(configuration.SOC_REQUEST,list):
        logging.error(f'SOC_REQUEST ist nicht vom Typ <list>')
        sys.exit()
    if ( l := len(configuration.SOC_REQUEST) ) != 8:
        logging.error(f'SOC_REQUEST hat {l} statt 8 ELemente.')
        sys.exit()
    for i in range(l):
        v = configuration.SOC_REQUEST[i]
        if not isinstance(v, int):
            logging.error(f'Element {i} in SOC_REQUEST ist nicht vom Typ int.')
            sys.exit()
        if i>0 and (v<0 or v>255):
            logging.error(f'Element {i} in SOC_REQUEST hat den Wert {v} und liegt nicht zwischen 0 und 255.')
            sys.exit()
            
    v = configuration.DST_SENDER_ID
    if not isinstance(v, int):
        logging.error(f'DST_SENDER_ID ist nicht vom Typ int.')
        sys.exit()
    if (v<0 or v>4095):
        logging.error(f'DST_SENDER_ID hat den Wert {v} und liegt nicht zwischen 0 und 4095.')
        sys.exit()
    v = configuration.DST_RECV_ID
    if not isinstance(v, int):
        logging.error(f'DST_RECV_ID ist nicht vom Typ int.')
        sys.exit()
    if ( v < 0 or v > (2**30-1) ):
        logging.error(f'DST_RECV_ID hat den Wert {v} und liegt nicht zwischen 0 und 2^30-1.')
        sys.exit()
    if not isinstance(configuration.DST_REQUEST,list):
        logging.error(f'DST_REQUEST ist nicht vom Typ <list>')
        sys.exit()
    if ( l := len(configuration.DST_REQUEST) ) != 8:
        logging.error(f'DST_REQUEST hat {l} statt 8 ELemente.')
        sys.exit()
    for i in range(l):
        v = configuration.DST_REQUEST[i]
        if not isinstance(v, int):
            logging.error(f'Element {i} in DST_REQUEST ist nicht vom Typ int.')
            sys.exit()
        if (v<0 or v>255):
            logging.error(f'Element {i} in DST_REQUEST hat den Wert {v} und liegt nicht zwischen 0 und 255.')
            sys.exit()

    # Ohne Fehler enden
    return(0)

# Hilfsroutinen für Hauptprogramm    
def openLocalLogfile(path):
    # Versuche, Datei unter 'path' ergänzend zu öffnen. 
    # Ein Rückgabewert von >0 bedeutet, daß das Öffnen nicht funktioniert hat.
    global savefile
    try:
        if not exists(path):
            logging.info(f'{path} existiert nicht. Lege neu an.')
            savefile = open(path,'a')
            savefile.write('Datum, Kilometerstand, Energiemenge, Start-SOC (Auto), End-SOC (Wallbox)\n')
            savefile.flush()
        else:
            logging.info(f'Öffne existierende Logdatei {path}.')
            savefile = open(path,'a')
        return 0
    except:
        logging.error(f'Konnte Ausgabedatei {path} nicht anlegen oder öffnen. Warte auf Änderung von configuration.py')
        return 1

def closeLocalLogfile():
    global savefile
    savefile.close()

def messageFromData(id,data):
    # Baue eine json-Nachricht zur Abfrage. Momentan wird ein extended-Falg gesetzt,
    # wenn die ID nicht in 11 Bit darstellbar ist. FIXME
    if id>4095:
        ext = 'true'
    else:
        ext = 'false'
    message = '{ "bus": "0", "type": "tx", "frame": [{ "id": '+str(id)+\
                 ', "dlc": 8, "rtr": false, "extd": '+ext+', "data": '+str(data)+' }] }'
    return message

#
# Callback-Funktionen
#

# Konfigurationsdatei geändert, Logdatei und Verbindungen schließen, Konfiguration neu einlesen.
def cb_cfChanged(event):
    global client
    logging.info(f'configuration.py geändert, starte Neueinlesen.')
    logging.debug(f'Schließe lokale Log-Datei {configuration.CHARGELOG_PATH}')
    closeLocalLogfile()
    logging.debug(f'Schließe MQTT-Verbindung zu {configuration.OPENWB_IP}')
    client.disconnect()
    logging.info(f'Lade configuration.py neu ein.')
    importlib.invalidate_caches()
    importlib.reload(configuration)
    logging.info(f'Prüfe configuration.py.')
    checkConfig()
#    ladevorgangclass.reloadConfig()
    # (ggf geänderte) Logdatei wird in Hauptschleife geöffnet
    # (ggf geänderte) MQTT-Verbindung wird in Hauptschleife hergestellt

# Erfolgreicher Verbindung zur OpenWB (CONNACK)
def on_connect(client, userdata, flags, rc):
    logging.info(f'Verbindung hergestellt zu {configuration.OPENWB_IP} mit Resultat {rc}')
    # Subscribing in on_connect() means that if we lose the connection and
    # reconnect then subscriptions will be renewed.
    client.subscribe('#')

def on_disconnect(client, userdata, rc):
    logging.info(f'Verbindung zu {configuration.OPENWB_IP} gelöst.')

# Energiezählerübertragung. In der Regel so hochfrequent, daß dem Empfang von Kilometerstand / SOC immer
# ein aktueller Zählerstand zugeordnet werden kann.
def cb_energycounter(client, userdata, msg):
    ladevorgang.setCounter(float(msg.payload))

# Steckerzustandes des Ladekabels.
def cb_plug(client, userdata, msg):
    if (b'true' in msg.payload) or (b'1' in msg.payload):
        ladevorgang.setPlug(1)
    else:
        ladevorgang.setPlug(0)

# Berechneter SOC aus der Wallbox
def cb_openwbsoc(client, userdata, msg):
    ladevorgang.setOpenwbsoc(msg.payload)
        
# Callback-Funktion für WiCAN-Statusbotschaften. Wenn ein status:online vom wican empfangen wird, wird
# unmittelbar danach eine SOC- und eine ODO-Abfrage geschickt.
def cb_status(client, userdata, msg):
    logging.info(f'WiCAN-Status: {msg.topic} {msg.payload}')
    # Statusbotschaft empfangen. Prüfen, ob online
    try:
        s=json.loads(msg.payload)
        if s['status'] == 'online':
            # Wenn das Fahrzeug online geht, kann der SOC und der Gesamtkm-Stand abgefragt werden
            logging.info(f'Fahrzeug ist online. Sende SOC- und DST-Anforderung')

            message = messageFromData(configuration.SOC_RECV_ID, configuration.SOC_REQUEST[0:8])
            logging.info(f'Sende SOC-Anforderung: {message}')
            client.publish(wican_topicTx, message)

            message = messageFromData(configuration.DST_RECV_ID, configuration.DST_REQUEST[0:8])
            logging.info(f'Sende DST-Anforderung: {message}')
            client.publish(wican_topicTx, message)
        else:
            logging.info(f'Fahrzeugstatus ist nicht <<online>>')
    except Exception as e:
        logging.warning(f'Fehler beim Parsen der Statusnachricht: {e}')

# Callback-Funktion für eine CAN-Rx-Botschaft. Prüft, ob es eine SOC- oder ODO-haltige Botschaft ist
# und sendet im Fall eines SOC den Wert zur OpenWB. Mehrteilige Botschaften werden aneinander gehängt.
def cb_rx(client, userdata, msg):
    # CAN-Botschaft empfangen. Prüfen, ob SOC und dann auswerten und and openWB pushen
    #
    # Es können alle möglichen Botschaften eintreffen. Gültig sind die, die mit
    # configuration.VALID_SOC_ANSWER und configuration.VALID_DST_ASWER definiert sind.
    # Für mehrteilige Botschaften muß ein Speicher bereitstehen, der 
    logging.debug(f'Empfangene CAN-Botschaft: {msg.payload}')
    try:
        frame = json.loads(msg.payload)['frame'][0]
        # json.loads(msg.payload): String der Nutzlast als dict
        # json.loads(msg.payload)['frame']: Datenframe in f ergibt eine Liste mit einem Element
        # json.loads(msg.payload)['frame'][0]: Das nullte und einzige Element der Liste ist ein dict
        id = frame['id']	# json.loads(msg.payload)['frame'][0]['id']: Sender-ID
        data = frame['data']    # json.loads(msg.payload)['frame'][0]['data']: Liste der Nutzbytes vom CAN
    except Exception as e:
        logging.error(f'Fehler beim json-Parsen der empfangenen CAN-Botschaft: {e}')

    # Prüfe, ob die Botschaft mehrteilig oder einteilig ist
    tpType =  data[0] // 16        # Oberen Nibble von Byte 0 extrahieren
    cb_rx.messageComplete = False
    if tpType == 1:
        # Erster Teil einer mehrteiligen Botschaft
        # Anzahl zu empfangenen Bytes: 12 Bit aus unterem Nibble Byte 0 und ganzes Byte 1
        cb_rx.bytesToReceive = (data[0] & 15)*256 + data[1]
        cb_rx.payload = [id]
        cb_rx.payload.extend(data[2:8])
        cb_rx.bytesReceived = 6
        logging.debug(f'Ersten Teil einer mehrteiligen Botschaft empfangen: {cb_rx.payload}')
        # zu Sende-ID des Senders die passende Empfänger-ID suchen:
        if id == configuration.DST_SENDER_ID:
            flowCtrl = messageFromData(configuration.DST_RECV_ID, [48,0,100,170,170,170,170,170])
        elif id == configuration.SOC_SENDER_ID:
            flowCtrl = messageFromData(configuration.SOC_RECV_ID, [48,0,100,170,170,170,170,170])
        # Fordere alle weitere Botschaften an mit 100ms Pause zwischen den Frames
        logging.debug(f'Aufforderung für Folgeteile absetzen: {flowCtrl}')
        client.publish(wican_topicTx, flowCtrl)
    elif tpType == 2:
        # Botschaft ist ein Folgeteil einer mehrteiligen Botschaft. Der Index sollte im unteren Nibble von Byte 0 stehen
        # hier wird einfach gehofft, daß die Botschaften in der richtigen Reihenfolge ankommen. Sie werden stumpf angehängt.
        cb_rx.payload.extend(data[1:8])     # Je Nachfolger sollten 7 Nutzbytes kommen
        cb_rx.bytesReceived += 7
        if cb_rx.bytesReceived >= cb_rx.bytesToReceive:
            cb_rx.messageComplete = True
        logging.debug(f'Mehrteilige Botschaft komplett: {cb_rx.payload}')
    elif tpType == 0:
        # Einteilige Botschaft
        cb_rx.payload = [id]
        cb_rx.payload.extend(data[1:8])
        cb_rx.bytesReceived = 7
        cb_rx.messageComplete = True
        logging.debug(f'Einteilige Botschaft: {cb_rx.payload}')
    else:
        logging.warning('Botschaft mit unbekanntem CAN-TP-Botschaftstyp oder FlowControl empfangen.')

    if cb_rx.messageComplete:
        #in payload liegt eine Liste der komplett empfangenen Botschaft vor
        # Erwartungswerte zusammenbauen
        lenSOC = configuration.SOC_REQUEST[0]
        expectSOC = configuration.SOC_REQUEST[1:1+lenSOC]
        expectSOC[0] += 64
        logging.debug(f'lenSOC: {lenSOC}; expectSOC: {expectSOC}')
        lenDST = configuration.DST_REQUEST[0]
        expectDST = configuration.DST_REQUEST[1:1+lenDST]
        expectDST[0] += 64
        logging.debug(f'lenDST: {lenDST}; expectDST: {expectDST}')
        if cb_rx.payload[0] == configuration.SOC_SENDER_ID and cb_rx.payload[1:1+lenSOC] == expectSOC:
            soc=configuration.getSoc(cb_rx.payload)
            logging.info(f'Fahrzeug-SOC ist {soc}')
            ladevorgang.setSoc(soc)
            logging.debug(f'SOC-Wert von {soc} an {configuration.OPENWB_SETSOC_TOPIC} schicken.')
            try:
                client.publish(configuration.OPENWB_SETSOC_TOPIC, soc)     #SOC-Wert an die OpenWB schicken
            except Exception as e:
                logging.error(f'Schreiben des SOC an die Wallbox ist fehlgeschlagen: {e}')
        elif cb_rx.payload[0] == configuration.DST_SENDER_ID and cb_rx.payload[1:1+lenDST] == expectDST:
            odo=configuration.getOdo(cb_rx.payload)
            ladevorgang.setOdo(odo)
            logging.info(f'Fahrzeug-Kilometerstand ist {odo}')
        else:
            logging.warning(f'Empfangene Botschaft: {cb_rx.payload} ist keine gültige Antwort auf eine konfigurierte Anfrage')

#
# Hauptprogramm
#
FORMAT = "%(asctime)s;%(levelname)9s;[%(filename)19s:%(lineno)3s - %(funcName)16s() ] %(message)s"
logging.basicConfig(encoding='utf-8', format=FORMAT, level=logging.getLevelName(configuration.LOGLEVEL))
logging.critical('Starte soc_helper Version 2024-04-28')

#Konfiguration prüfen
checkConfig()
# Globale Klasse für Ladedaten erstellen
ladevorgang = ladevorgangclass.ladevorgangclass()
# Überwachung der Konfigurationsdatei einrichten, damit sie bei Änderungen neu eingelesen wird
patterns = ["configuration.py"]
ignore_patterns = None
ignore_directories = True
case_sensitive = True
cfHandler = PatternMatchingEventHandler(patterns, ignore_patterns, ignore_directories, case_sensitive)
cfHandler.on_modified = cb_cfChanged
cfWatchdog = Observer()
cfWatchdog.schedule(cfHandler, ".", recursive = False)
cfWatchdog.start()
# MQTT-Client einrichten und mit dem Broker verbinden
client = mqtt.Client()
client.on_pre_connect = None
client.on_connect = on_connect
client.on_subscribe = None
client.on_disconnect = on_disconnect

while True:
    # Die Schleife ist nötig, da beim Neueinlesen von configuration.py eine bestehende MQTT-Verbindung geschlossen wird
    # (Schließlich könnte eine andere Zieladresse konfiguriert worden sein.). Nach Beenden der Verbindung endet
    # client.loop_forever()

    # Datei zum Abspeichern der Ladedaten öffnen (könnte sich durch Konfigurationsänderung geändert haben)
    # Momentan terminiert das Programm in diesem Fall.
    # Versuche, die lokale Logdatei zu öffnen und verbleibe in einer Schleife, bis das klappt
    # (die Überwachung von configuration.py sollte bei Dateiänderungen ein Neueinlesen verursachen)
    rv = openLocalLogfile(configuration.CHARGELOG_PATH)
    if rv > 0:
        #Öffnen fehlgeschlagen. Endlosschleife, die durch Änderung von configuration.py unterbrochen wird:
        while True:
            time.sleep(1)
    # Öffnen erfolgreich, Filehandle-Übergabe an Klasse
    ladevorgang.setLogfilehandle(savefile)

    # Verbindung zum MQTT-Broker in der Wallbox herstellen
    # Verbesserungsidee: Durch Änderung der Konfiguration ist es möglich, daß eine ungültige IP-Adresse anliegt.
    #		Dann sollte hier eine Logmeldung ausgegeben werden und gewartet werden, bis configuration sich erneut 
    #		verändert hat.
    logging.debug(f'Verbindungsversuch mit MQTT-Broker der Wallbox unter Adresse {configuration.OPENWB_IP}.')
    try:
        # MQTT-Topic-Abos einrichten und mit Callback-Funktionen verbinden
        client.message_callback_add(configuration.OPENWB_CHARGECOUNTER_TOPIC, cb_energycounter)
        client.message_callback_add(configuration.OPENWB_GETPLUGSTATE_TOPIC, cb_plug)
        client.message_callback_add(configuration.OPENWB_GETSOC_TOPIC, cb_openwbsoc)
        wican_topicStatus = 'wican/'+str(configuration.WICAN_SER)+'/status/#'
        logging.debug(f'WiCAN-Status-Topic: {wican_topicStatus}')
        wican_topicRx = 'wican/'+str(configuration.WICAN_SER)+'/can/rx/#'
        logging.debug(f'WiCAN-Empfangs-Topic: {wican_topicRx}')
        wican_topicTx = 'wican/'+str(configuration.WICAN_SER)+'/can/tx'
        logging.debug(f'WiCAN-Sende-Topic: {wican_topicTx}')
        client.message_callback_add(wican_topicStatus, cb_status)
        client.message_callback_add(wican_topicRx, cb_rx)
        client.connect(configuration.OPENWB_IP, 1883, 60)
    except Exception as e:
        logging.error(f'FATAL: Fehler beim Verbinden zum Broker auf {configuration.OPENWB_IP}: {e}')
        quit()
    client.loop_forever()
    time.sleep(1)	# Nach Ende der Verbindung zum MQTT-Broker der Wallbox kurz warten vor neuer Kontaktaufnahme, sonst Fehler
    
